# Originally published as AnsiColors
# Copyright (c) 2012 Giorgos Verigakis <verigak@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# This module is a modified version of the original AnsiColors module, and further
# modified by github.com/brandoncript to be used in the Tinta project. It is
# distributed under the same license as the original module, as well as the
# MIT License.

import re
from typing import (
    Any,
    cast,
    List,
    Optional,
    overload,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from .constants import ANSI_RESET, ANSI_RESET_OCT, ANSI_STYLES, ANSI_STYLES_OFF, SEP
from .typ import InvalidStyleError

if TYPE_CHECKING:
    from .tinta import Tinta


def was_reset(s: str) -> bool:
    return s.strip().endswith(ANSI_RESET) or s.endswith(ANSI_RESET_OCT)


def ensure_reset(s: str, np: "Union[Tinta.Part, None]" = None) -> str:
    if was_reset(s) or (np and np.has_formatting):
        return s
    last_char_idx = len(s.rstrip())
    return f"{s[:last_char_idx]}{ANSI_RESET}{s[last_char_idx:]}"


def _parse_rgb(s: str) -> Tuple[int, ...]:
    if not isinstance(s, str):
        raise ValueError(f"Could not parse color '{s}'")
    s = s.strip().replace(" ", "").lower()

    # 6-digit hex
    match = re.match("#([a-f0-9]{6})$", s)
    if match:
        core = match.group(1)
        return tuple(int(core[i : i + 2], 16) for i in range(0, 6, 2))

    # 3-digit hex
    match = re.match("#([a-f0-9]{3})$", s)
    if match:
        return tuple(int(c * 2, 16) for c in match.group(1))

    # rgb(x,y,z)
    match = re.match(r"rgb\((\d+,\d+,\d+)\)", s)
    if match:
        return tuple(int(v) for v in match.group(1).split(","))

    raise ValueError(f"Could not parse color '{s}'")


def _join(*values: Union[int, str]) -> str:
    """
    Join a series of values with semicolons. The values
    are either integers or strings, so stringify each for
    good measure. Worth breaking out as its own function
    because semicolon-joined lists are core to ANSI coding.
    """
    return ";".join(str(v) for v in values).strip(";").strip()


def _color_code(
    spec: Union[str, int, Tuple[int, int, int], List[int]], base: int
) -> str:
    """Workhorse of encoding a color. Give preference to named colors from
    ANSI, then to specific numeric or tuple specs. If those don't work,
    try looking up look CSS color names or parsing CSS hex and rgb color
    specifications.

    Args:
        spec: Unparsed color specification
        base (int): Either 30 or 40, signifying the base value
                    for color encoding (foreground and background respectively).
                    Low values are added directly to the base. Higher values use `
                    base + 8` (i.e. 38 or 48) then extended codes.

    Returns:
        Discovered ANSI color encoding.

    Raises:
        ValueError if cannot parse the color spec.
    """
    if isinstance(spec, str):
        spec = spec.strip().lower()

    if spec == "default" or spec == 0:
        return "0"
    elif isinstance(spec, int) and 1 <= spec <= 255:
        return _join(base + 8, 5, spec)
    elif isinstance(spec, (tuple, list)):
        return _join(base + 8, 2, _join(*spec))
    elif str(spec).startswith("#"):
        rgb = _parse_rgb(str(spec))
        # parse_rgb raises ValueError if cannot parse spec
        return _join(base + 8, 2, _join(*rgb))
    else:
        raise ValueError(f"Could not parse color '{spec}'")


def _style_name(code: int) -> str:
    try:
        return ANSI_STYLES[code]
    except IndexError:
        raise InvalidStyleError(f"Invalid style code '{code}'")


def _style_codes(style: Union[str, int]) -> Tuple[int, int]:
    if isinstance(style, str):
        style = style.lower()
    on = ANSI_STYLES.index(style) + 1
    off = next((c for i, c in ANSI_STYLES_OFF for s in i if s == style), 0)
    return on, off


def ansi_styles(style: Union[str, int]) -> Tuple[str, str]:
    """Returns ANSI on and off strings for a style name or code.

    Args:
        style (str | int): A style name or code.

    Returns:
        Tuple[str, str]: The on and off ANSI codes for the style.
    """
    on, off = _style_codes(style)
    return f"\x1b[{on}m", f"\x1b[{off}m"


def validate_styles(*styles: Union[int, str]) -> Tuple[str, ...]:
    """Validates a list of style names. If a style is not found in the ANSI
    styles, it is removed from the list.
    """

    _styles: List[Union[int, str]] = list(styles)

    if next((isinstance(s, int) for s in styles), False):
        _styles = [_style_name(int(s)) for s in _styles]

    _styles = [str(s).lower() for s in _styles]

    if not all([s in ANSI_STYLES for s in _styles]):
        available_styles = ", ".join([f"'{s}'" for s in ANSI_STYLES])
        err = f"Invalid style string '{_styles}', must be one of {available_styles}."
        raise InvalidStyleError(err)

    return cast(Tuple[str, ...], tuple(_styles))


@overload
def tint(
    instance: "Tinta", *s: Any, color: Union[str, int], sep: str = SEP
) -> "Tinta": ...


@overload
def tint(
    instance: "Tinta", color: Union[str, int], *s: Any, sep: str = SEP
) -> "Tinta": ...


def tint(instance: "Tinta", *args, **kwargs) -> "Tinta":
    """Adds segments of text colored with the specified color.
    Can be used in place of calling named color methods.

    Args:
        instance (Tinta): The Tinta instance.
        color (str | int, optional): A color name or ANSI color index. Defaults to first argument.
        *s: Any: Segments of text to add.
        sep (str, optional): Used to join strings. Defaults to ' '.

    Returns:
        self
    """
    from .tinta import Tinta

    if not isinstance(instance, Tinta):
        raise AttributeError("tint() must be called from a Tinta instance.")

    self = instance

    # check if the first argument is a known color or valid ANSI code, or comes from kwargs
    s = args
    color = kwargs.get("color", None)
    sep = kwargs.get("sep", SEP)
    if color is None:
        if not len(s) > 1:
            raise AttributeError(
                "If no color is specified, tint() requires at least two arguments."
            )
        color = s[0]
        s = s[1:]

    self.set_color(color)
    self.push(*s, sep=sep)
    return self


def stylize(
    s: Optional[str],
    lp: "Union[Tinta.Part, None]",
    mp: "Tinta.Part",
    np: "Union[Tinta.Part, None]",
) -> str:
    """Wraps a string with ANSI styles.

    Args:
        s (str): The string to format - may be plain or pre-formatted.
        lp (Tinta.Part): The previous part.
        mp (Tinta.Part): The current (middle) part.
        np (Tinta.Part): The next part.
    """

    if s is None:
        s = ""

    if not mp:
        return s

    # On
    on = ""
    if mp.has_formatting:
        if not lp or not lp.has_formatting:
            on = _join(*[_style_codes(st)[0] for st in mp.style])
        else:
            if mp.style != lp.style:
                on = _join(
                    *[_style_codes(st)[0] for st in mp.style if st not in lp.style]
                )
        if mp.color_code and (not lp or (mp.color_code != lp.color_code)):
            on = _join(on, _color_code(mp.color_code, 30))

    # Off
    off = ""
    if mp.has_formatting:
        if (
            not np
            or not np.has_formatting
            or mp.styler._force_clear
            or bool(mp.style and not np.style and mp.color_code != np.color_code)
        ):
            off = "0"
        elif mp.style != np.style:
            styles_not_in_n = [s for s in mp.style if s not in np.style]
            off = _join(
                *list(sorted(set([_style_codes(st)[1] for st in styles_not_in_n])))
            )

    # Apply formatting
    left = f"\x1b[{on}m" if on else ""
    right = f"\x1b[{off}m" if off else ""

    last_char_idx = len(s.rstrip())

    return f"{left}{s[:last_char_idx]}{right}{s[last_char_idx:]}"


def is_ansi_str(s: str) -> bool:
    return s.startswith("\x1b[") or s.startswith("\033[")


def ansi_color_to_int(s: str) -> int:
    """Converts an ANSI escape sequence to a base 10 integer. Accepts both octal and decimal sequences, and from 1 to 3 segments separated by semicolons."""

    if not is_ansi_str(s):
        raise ValueError(f"String '{s}' is not an ANSI escape sequence.")

    if s.startswith("\x1b["):
        s = s[2:]
    elif s.startswith("\033["):
        s = s[3:]

    if s.endswith("m"):
        s = s[:-1]

    seg = s.split(";")
    if len(seg) >= 3:
        ch = seg[-3]
        if ch == "38" or ch == "48":
            return int(seg[-1])
        return int(seg[-1])

    raise ValueError(f"Could not parse ANSI escape sequence '{s}'")



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""This class is a low-level helper class for managing colors in Tinta."""

import configparser
import sys
from pathlib import Path
from typing import Dict, List, Optional, Type, Union

from .stylize import ansi_color_to_int, is_ansi_str
from .typ import MissingColorError

colors_ini = configparser.ConfigParser()


def _alias_keys(colors: "Union[AnsiColors, Type[AnsiColors]]", search: str, repl: str):
    """Sets up an alias key for a color."""

    for k in (k for k in colors_ini["colors"].keys() if search.lower() in k.lower()):
        alias_key = k.replace(search.lower(), repl.lower())
        if (
            alias_key not in colors_ini["colors"]
            and alias_key not in colors._color_dict
        ):
            colors._color_dict[alias_key] = int(colors_ini["colors"][k])
            colors_ini["colors"][alias_key] = colors_ini["colors"][k]


def _check_path(path: Optional[Union[str, Path]] = None):
    """Loads colors from a file."""
    path = Path(path) if path else Path(__file__).parent / "colors.ini"
    if not path.is_absolute():
        path = Path().cwd() / path
    if not path.exists():
        raise FileNotFoundError(
            f"Tinta failed to load colors, '{path}' does not exist."
        )

    # if path is a dir, look for colors.ini
    if path.is_dir():
        path = path / "colors.ini"

    # if there is still no colors.ini file, look for one in Path.cwd() or PYTHONPATH
    if not path.exists():
        for p in [Path().cwd(), Path(sys.path[0])]:
            if (p / "colors.ini").exists():
                path = p / "colors.ini"
                break

    if not path.exists():
        raise FileNotFoundError(
            f"Tinta failed to load colors, could not find 'colors.ini' in cwd or in PYTHONPATH. Please provide a valid path to a colors.ini file."
        )
    return path


class AnsiColors:
    """Color builder for Tinta's console output.

    ANSI color map for console output. Get a list of colors here =
    http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors

    Or run Tinta.discover() to see all 256 colors on your system.

    You can change the colors the terminal outputs by changing the
    ANSI values in colors.ini.
    """

    _initialized = False
    _colors_ini_path = None
    _color_dict: Dict[str, int] = {}

    def __init__(self, path: Optional[Union[str, Path]] = None):

        if not AnsiColors._initialized:
            path = _check_path(path)
            AnsiColors.load_colors(path)
            AnsiColors._colors_ini_path = path
            AnsiColors._initialized = True

    @classmethod
    def load_colors(cls, path: Union[str, Path]):
        """Loads colors from a file."""

        colors_ini["colors"] = {}
        colors_ini.read(path)
        cls._color_dict = {k: int(v) for (k, v) in colors_ini["colors"].items()}
        # Check if any of the color names loaded match the built-in methods. If so, raise an error.
        # for col in loaded_colors.keys():
        #     if hasattr(cls, col):
        #         # and not str(getattr(cls, col)).startswith(
        #         #     "functools.partial(<bound method Tinta.tint"
        #         # ):
        #         # if the color is a 'tint' method, update it
        #         raise AttributeError(
        #             f"Cannot overwrite built-in method '{col}' with color name. Please rename the color in '{path}'."
        #         )

        # # Add the loaded colors class
        # for k, v in colors_ini["colors"].items():
        #     setattr(cls, k, int(v))

        _alias_keys(cls, "gray", "grey")
        _alias_keys(cls, "grey", "gray")

    def get(self, color: str) -> int:
        """Returns the ANSI code for a color.

        Args:
            color (str): A color name.

        Returns:
            int: The ANSI code for the color.
        """

        if color == "default":
            return 0

        if is_ansi_str(color):
            return ansi_color_to_int(color)

        if color not in self._color_dict:
            raise MissingColorError(f"Color '{color}' not found in colors.ini.")

        return self._color_dict[color]

    def reverse_get(self, code: int, ignore_errors: bool = False) -> str:
        """Returns the color name for an ANSI code.

        Args:
            code (int): An ANSI code.
            ignore_errors (bool, optional): If True, will return None if the code is not found. Defaults to False.

        Returns:
            str: The color name for the code.
        """

        if code == 0:
            return "default"

        for k, v in self._color_dict.items():
            if v == code:
                return k

        if not ignore_errors:
            raise MissingColorError(
                f"Color with ANSI code '{code}' not found in colors.ini."
            )

        return "[undefined color]"

    @property
    def color_list(self) -> List[str]:
        """Returns a list of all colors in the colors.ini file."""
        return list(self.color_dict.keys())

    @property
    def color_dict(self) -> Dict[str, int]:
        """Returns a dictionary of all colors in the colors.ini file."""
        if not self._color_dict:
            self._color_dict = {k: int(v) for (k, v) in colors_ini["colors"].items()}
        return self._color_dict



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""Tinta is a magical console output tool with support for printing in beautiful
colors and with rich formatting, like bold and underline. It's so pretty,
it's almost like a unicorn.
"""

import configparser
import functools
import os
import re
import sys
from pathlib import Path
from typing import Any, List, Optional, overload, Sequence, Union

from .ansi import AnsiColors
from .constants import (
    ANSI_STYLES,
    CURSOR_UP_ONE,
    ERASE_LINE,
    PREFER_PLAINTEXT,
    PUNC,
    SEP,
    SMART_FIX_PUNCTUATION,
    STEALTH,
)
from .discover import discover as _discover
from .stylize import _join, ensure_reset, is_ansi_str, stylize, tint
from .typ import copy_kwargs, MissingColorError, StringType

config = configparser.ConfigParser()


class _MetaTinta(type):

    _initialized = False
    _colors: AnsiColors

    def __init__(cls, name, bases, dct):
        super(_MetaTinta, cls).__init__(name, bases, dct)

        if not cls._initialized:
            cls._colors = AnsiColors()
            cls._initialized = True

    def load_colors(cls, path: Union[str, Path]):
        cls._initialized = False
        AnsiColors._initialized = False
        cls._colors = AnsiColors(path)


class Tinta(metaclass=_MetaTinta):
    """Tinta is a magical console output tool with support for printing in
    beautiful colors and with rich formatting, like bold and underline. It's
    so pretty, it's almost like a unicorn.

    All public methods chain together to form a builder pattern, e.g.:

    (Tinta('Some plain text')
        .white(' white')
        .blue(' blue')
        .red(' red')
        .bold().red(' bold red')
        .dark_gray()
        .dim(' dim').print())

    Args:
        *s (tuple(Any)): A sequence of one or more text strings, to be joined together.
        sep (str): Used to join segment strings. Defaults to ' '.

    Attributes:
        color (str, int):           A color string or ansi color code (int),
                                    e.g. 'white' or 42
        style (str):                A style string, e.g. 'bold', 'dim', 'underline'.
                                    Multiple styles are joined with a +
        parts (list):               A list of Part objects.
        get_parts(style: str):      Returns a list of string parts, or a list of Parts.

    Methods:
        to_str() -> str:           Returns a compiled rich text string, or use plaintext=True for plaintext.
        push() -> self:             Adds segments to this Tinta instance.
        pop(qty: int) -> self:              Removes the last 'qty' segments from this Tinta instance.
        bold() -> self:             Sets segments to bold.
        underline() -> self:        Sets segments to underline.
        dim() -> self:              Sets segments to a darker, dimmed color.
        strikethrough() -> self:    Sets segments to strikethrough.
        normal() -> self:           Removes all styles.
        clear() -> self:            Removes all styles and colors.
        line() -> self:             Adds segments on a new line.
        print():                    Prints the output of a Tinta instance, then clears.
    """

    def __init__(
        self,
        *s: Any,
        color: Union[str, int] = 0,
        styles: List[str] = [],
        sep: str = SEP,
    ):
        """Main intializer for Tinta

        Args:
            *s: Segments of text to add.
            color (Union[str, int], optional): A color name or ANSI color index. Defaults to None.
            styles (List[str], optional): A list of style strings. Defaults to None.
            sep (str, optional): Used to join strings. Defaults to ' '.
        """

        # Inject ANSI helper functions
        for c in Tinta._colors.color_dict:
            if hasattr(self, c) and not str(getattr(self, c)).startswith(
                "functools.partial(<bound method Tinta.tint"
            ):
                raise AttributeError(
                    f"Cannot overwrite built-in method '{c}' with color name. Please rename the color in '{Tinta._colors._colors_ini_path}'."
                )
            setattr(self, c, functools.partial(self.tint, color=c))

        self._styler = Tinta.Styler(color=color, styles=styles)
        self._parts: List["Tinta.Part"] = []
        self._prefixes: List[str] = []

        if s:
            self.push(*s, sep=sep)

    def __call__(self, *s: Any, sep: str = SEP) -> "Tinta":
        return self.push(*s, sep=sep)

    def __repr__(self) -> str:
        """Generates a string representation of the current
        Tinta instance.

        Returns:
            str: Plaintext string
        """
        return str(self.to_str(plaintext=True))

    def __str__(self) -> str:
        """Generates a string representation of the current
        Tinta instance, colorized.

        Returns:
            str: Colorized string"""

        return self.to_str()

    def _get_sep(
        self,
        mp: "Tinta.Part",
        np: Optional["Tinta.Part"],
        sep: Optional[str],
    ) -> str:
        if not np:
            return ""
        return sep if sep is not None else mp.sep

    def _get_str(
        self,
        attr: StringType,
        lp: Optional["Tinta.Part"],
        mp: "Tinta.Part",
        np: Optional["Tinta.Part"],
        sep: Optional[str] = None,
        fix_punc: bool = SMART_FIX_PUNCTUATION,
    ) -> str:

        s: str = (
            mp.pln
            if attr == "pln"
            else mp.fmt(lp, np) if attr == "fmt" else mp.esc(lp, np)
        )

        if fix_punc:
            # next char is empty
            if not np or not np.pln:
                return s

            # current char is empty
            if not mp.pln:
                return s

            # next or last char is newline
            if np.pln.startswith(os.linesep) or mp.pln.endswith(os.linesep):
                return s

            # next char is punctuation and affects the separator
            if np.pln[0] in PUNC and (len(np.pln) == 1 or np.pln[1] == " "):
                return s

        return f"{s}{self._get_sep(mp, np, sep)}"

    def to_str(
        self,
        sep: Optional[str] = None,
        plaintext: bool = False,
        escape_ansi: bool = False,
        fix_punctuation: bool = SMART_FIX_PUNCTUATION,
    ) -> str:
        """Returns a compiled rich text string, joined by 'sep'. Note that any separators included with a
        part will be returned as part of the string - adding a separator here may result in a double separator.

        Args:
            sep (str, optional): Used to join strings. Defaults to the separator used when the part was added, or ' '. Setting this will override the part's separator.
            plaintext (bool, optional): If True, returns a plaintext string. Defaults to False.
            escape_ansi (bool, optional): If True, will escape ANSI codes' \\ → \\\\. Defaults to False.
            fix_punctuation (bool, optional): If True, will fix punctuation spacing. Defaults to SMART_FIX_PUNCTUATION or True.

        Returns:
            str: A rich text string.
        """

        if not self._parts:
            return ""

        attr: StringType = "pln" if plaintext else "esc" if escape_ansi else "fmt"

        return "".join(
            [
                self._get_str(attr, lp, p, np, sep, fix_punctuation)
                for lp, p, np in self._parts_tuple
            ]
        )

    @property
    def color(self) -> str:
        """A color string, e.g. 'white' or 'blue'.

        Returns:
            str: A color string or ansi color code (int)."""
        return self._styler.color

    def set_color(self, value: Union[int, str]):
        """Sets the current color of the Tinta instance.

        Args:
            value (Union[int, str]): A color string or ansi color code (int),
            e.g. 'white' or 42.
        """

        self._styler.set_color(value)

    @property
    def color_code(self) -> int:
        """The ANSI code for the current color.

        Returns:
            int: The ANSI code for the current color."""
        return self._styler.color_code or 0

    @property
    def style(self) -> List[str]:
        """A style string, e.g. 'bold', 'dim', 'underline'.

        Returns:
            List[str]: A list of style strings."""
        return self._styler.active_styles or []

    @style.setter
    def style(self, value: Union[Sequence[str], Sequence[int]]):
        self._styler.set_styles(value)

    @property
    def parts(self) -> List["Tinta.Part"]:
        """A list of Tinta.Part objects.

        Returns:
            List[Part]: A list of Tinta.Part objects"""

        return self._parts

    @property
    def _parts_tuple(self):

        for i, p in enumerate(self._parts):
            lp = self._parts[i - 1] if i > 0 else None
            np = self._parts[i + 1] if i < len(self._parts) - 1 else None
            yield lp, p, np

    @property
    def current_part(self) -> "Tinta.Part":
        if not self._parts:
            self._parts.append(Tinta.Part("", Tinta.Styler(), sep=SEP))
        return self._parts[-1]

    @property
    def has_formatting(self) -> bool:
        """Returns True if any part has formatting.

        Returns:
            bool: True if any part has formatting."""
        return bool(next((p.has_formatting for p in self._parts), False))

    @property
    def parts_fmt(self) -> list:
        """Returns a list of richly formated string parts

        Returns:
            str: A list of formatted strings."""
        return [p.fmt(lp, np) for lp, p, np in self._parts_tuple]

    @property
    def parts_pln(self) -> list:
        """Returns a list of plaintext string parts

        Returns:
            str: A list of plaintext strings."""
        return [part.pln for part in self._parts]

    @property
    def parts_esc(self) -> list:
        """Returns a list of escaped formatted string parts, so that special characters are visible (i.e., backslashes are doubled).

        Returns:
            str: A list of esc strings."""
        return [p.esc(lp, np) for lp, p, np in self._parts_tuple]

    def push(
        self,
        *s: Any,
        sep: Optional[str] = SEP,
    ) -> "Tinta":
        """Adds segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """

        # If an empty set of segments is passed, skip to the
        # next segment (to prevent duplicating whitespace).
        # if not s:
        #     return self

        if sep is None:
            sep = SEP

        # Join all s parts with the specified separator
        pln = "" if not s else sep.join([str(x) for x in s])

        # Collect any prefixes that may have been set
        if self._prefixes:
            pln = "".join(self._prefixes) + pln
            self._prefixes = []

        # TODO: Add support for background colors

        # If the current part's string is not empty, add a new part
        if self.current_part.s:
            self._parts.append(Tinta.Part(pln, self._styler.copy(), sep=sep))

        # Otherwise, update the current part with additional styles and current string
        elif not self.current_part.s:
            self.current_part.s = pln
            self.current_part.styler = self._styler.copy()
            self.current_part.sep = sep

        self._styler._force_clear = False

        return self

    def pop(self, qty: int = 1) -> "Tinta":
        """Removes the last 'qty' segments from this Tinta instance

        Args:
            qty (int, optional): Number of segments to remove. Defaults to 1.

        Returns:
            self
        """
        for _ in range(min(qty, len(self._parts))):
            self._parts.pop()
        self._styler = self._parts[-1].styler.copy() if self._parts else Tinta.Styler()
        return self

    @overload
    def tint(self, *s: Any, color: Union[str, int], sep: str = SEP) -> "Tinta": ...

    @overload
    def tint(self, color: Union[str, int], *s: Any, sep: str = SEP) -> "Tinta": ...

    def tint(self, *args, **kwargs) -> "Tinta":
        """Adds segments of text colored with the specified color.
        Can be used in place of calling named color methods.

        Args:
            instance (Tinta): The Tinta instance.
            color (str | int, optional): A color name or ANSI color index. Defaults to first argument.
            *s: Any: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """

        return tint(self, *args, **kwargs)

    @overload
    def inspect(self, code: int, name: None = None, throw: bool = False) -> str: ...

    @overload
    def inspect(
        self, code: None = None, name: str = "default", throw: bool = False
    ) -> int: ...

    def inspect(
        self,
        code: Optional[int] = None,
        name: Optional[str] = None,
        throw: bool = False,
    ) -> Union[str, int, None]:
        """Queries the known colors by code or name, and returns the corresponding name or code, respectively.

        Args:
            code (int, optional): An ANSI color code. Defaults to None.
            name (str, optional): A color name. Defaults to None.

        Returns:
            Union[str, int]: A color name or code, if found, otherwise None.
        """

        if code is not None and name is not None:
            raise AttributeError(
                "Tinta.inspect() accepts either a code or a name, not both."
            )

        try:
            if code is not None:
                if code == 0:
                    return "default"
                return self._colors.reverse_get(code)

            elif name is not None:
                if name == "default":
                    return 0
                return self._colors.get(name)
        except MissingColorError as e:
            if throw:
                raise e
        return None

    def bold(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds bold segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._styler.toggle_styles("bold")
        self.push(*s, sep=sep)
        return self

    @copy_kwargs(bold)
    def b(self, *args, **kwargs):
        return self.bold(*args, **kwargs)

    def underline(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds underline segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._styler.toggle_styles("underline")
        self.push(*s, sep=sep)
        return self

    @copy_kwargs(underline)
    def u(self, *args, **kwargs):
        return self.underline(*args, **kwargs)

    @copy_kwargs(underline)
    def _(self, *args, **kwargs):
        return self.underline(*args, **kwargs)

    def strikethrough(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds strikethrough segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._styler.toggle_styles("strikethrough")
        self.push(*s, sep=sep)
        return self

    def dim(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds darker (dimmed) segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._styler.toggle_styles("dim")
        self.push(*s, sep=sep)
        return self

    def normal(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Removes all styles, then adds segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        # self._prefixes.append('\033[24m\033[21m')
        self._styler.clear_styles()
        self.push(*s, sep=sep)
        return self

    def default(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Removes all styles, then adds segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        # self._prefixes.append('\033[24m\033[21m')
        self.set_color(0)
        self.push(*s, sep=sep)
        return self

    def clear(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Removes all styles and colors, then adds segments to this
        Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._styler.clear_all(force=True)
        self.push(*s, sep=sep)
        # self._prefixes.append('\033[24m\033[21m')
        return self

    def nl(self, *s, sep: str = SEP) -> "Tinta":
        """Adds segments to this Tinta instance, preceded by a new line.

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._prefixes = [os.linesep]
        self.push(*s, sep=sep)
        return self

    def print(
        self,
        sep: Optional[str] = None,
        end="\n",
        file=sys.stdout,
        flush=False,
        plaintext=False,
        force=False,
    ):
        """Prints a Tinta composite to the console. Once printed,
        this Tinta instance is cleared of all configuration, but can
        can continue to be used to print.

        Env: These environment variables, when set, affect Tinta globally.
            TINTA_STEALTH (not None): Hides all console output. Can be
                                      overridden by 'force'.
            TINTA_PLAINTEXT (not None): Prints all output in plaintext.

        Args:
            sep (str, optional): Used to join strings. Defaults to ' ', or whatever was used when the part was added. Setting this will override the part's separator.
            end (str, optional): String terminator. Defaults to '\n'.
            file (optional): File to write to. Defaults to sys.stdout.
            flush (bool, optional): Clears the current console line.
            plaintext (bool, optional): Prints in plaintext. Defaults to False.
            force (bool, option): Forces printing, overriding TINTA_STEALTH.
        """
        # We don't print if the TINTA_STEALTH env is set
        if STEALTH and not force:
            return

        use_plaintext = True if plaintext or PREFER_PLAINTEXT else False
        s = self.to_str(sep=sep, plaintext=use_plaintext)
        # if s.strip() does not end with reset, add it after the last non-whitespace character
        if not use_plaintext:
            s = ensure_reset(s)
        print(
            self.to_str(sep=sep, plaintext=use_plaintext),
            end=end,
            file=file,
            flush=flush,
        )

        self.clear()
        self._parts = []

    def __getattr__(self, name: str) -> "Tinta":
        """Returns a tinted segment of text.

        Args:
            name (str): Name of the color.

        Returns:
            Tinta: A Tinta instance.
        """

        if not name.startswith("_"):
            # if hasattr(self.colors, name):
            #     return cast("Tinta", self.tint(color=name))

            # else:
            try:
                return self.__getattribute__(name)  # type: ignore
            except AttributeError as e:
                known_colors = "\n - ".join(self._colors.color_list)
                known_colors = f" - {known_colors}"
                raise AttributeError(
                    f"'{name}' not found.\nDid you try and access a color that doesn't exist? Available colors:\n{known_colors}\n"
                ) from e

        return self.__getattribute__(name)  # type: ignore

    @staticmethod
    def discover(background=False):
        """Prints all 256 colors in a matrix on your system. If background is True,
        it will print background colors with numbers on top."""
        _discover(background)

    @staticmethod
    def clearline():
        """Clears the current printed line."""

        if sys.stdout:
            sys.stdout.write(CURSOR_UP_ONE)
            sys.stdout.write(ERASE_LINE)
            sys.stdout.flush()

    @staticmethod
    def up():
        """Moves up to the previous line."""

        if sys.stdout:
            sys.stdout.write(CURSOR_UP_ONE)
            sys.stdout.flush()

    class Styler:
        """A class to hold format and style config for a Tinta part."""

        color: str = "default"
        color_code: int = 0
        bold: bool = False
        dim: bool = False
        italic: bool = False
        underline: bool = False
        blink: bool = False
        rapid_blink: bool = False
        invert: bool = False
        conceal: bool = False
        strikethrough: bool = False
        _force_clear: bool = False

        def __init__(
            self,
            *,
            color: Union[str, int] = 0,
            styles: List[str] = [],
            force_clear=False,
        ):

            self.set_color(color)
            self.set_styles(styles)
            self._force_clear = force_clear

        def __repr__(self):
            color_str = f"{self.color}{{{self.color_code}}}"
            if not self.active_styles:
                return color_str
            if not any([self.color, self.active_styles]):
                return "(plain)"
            return " - ".join([color_str, _join(*self.active_styles)])

        def __str__(self):
            return self.__repr__()

        def __eq__(self, other):
            if not isinstance(other, Tinta.Styler):
                return False
            return (
                self.color == other.color and self.active_styles == other.active_styles
            )

        def copy(self):
            new = self.__new__(self.__class__)
            new.color = self.color
            new.color_code = self.color_code
            for k in ANSI_STYLES:
                setattr(new, k, getattr(self, k))
            return new

        @property
        def active_styles(self) -> List[str]:
            """Returns a list of active styles."""
            return [st for st in ANSI_STYLES if getattr(self, st, None)]

        @property
        def inactive_styles(self) -> List[str]:
            """Returns a list of inactive styles."""
            return [st for st in ANSI_STYLES if not getattr(self, st, None)]

        def set_color(self, color: Union[str, int]):

            if isinstance(color, str):
                code = Tinta._colors.get(color)
                self.color = (
                    Tinta._colors.reverse_get(code, ignore_errors=True)
                    if is_ansi_str(color)
                    else color
                )
                self.color_code = code
            else:
                self.color = Tinta._colors.reverse_get(color, ignore_errors=True)
                self.color_code = color

        def set_styles(self, styles: Union[Sequence[str], Sequence[int]]):

            from .stylize import validate_styles

            styles = validate_styles(*styles)

            for k in styles:
                setattr(self, k, True)

        def enable_styles(self, *styles: str):
            from .stylize import validate_styles

            styles = validate_styles(*styles)
            for style in styles:
                setattr(self, style, True)

        def disable_styles(self, *styles: str):
            from .stylize import validate_styles

            styles = validate_styles(*styles)
            for style in styles:
                setattr(self, style, False)

        def toggle_styles(self, *styles: str):
            from .stylize import validate_styles

            styles = validate_styles(*styles)
            for style in styles:
                setattr(self, style, not getattr(self, style))

        def clear_styles(self):
            for k in ANSI_STYLES:
                setattr(self, k, False)

        def clear_all(self, force: bool = False):
            for k in ANSI_STYLES:
                setattr(self, k, False)
            self.set_color(0)
            self._force_clear = force

    class Part:
        """A segment part of text, intended to be joined together
        for a print statement.

        Args:
            s (str):                    A segment of unstyled text.
            styler: (Tinta.Styler):     A Tinta.Styler object.
            sep (str):                  Used to join segment strings. Defaults to ' '.
        """

        def __init__(
            self,
            s: str,
            styler: "Optional[Tinta.Styler]" = None,
            sep: str = SEP,
        ):
            self.s = s
            self.styler = styler or Tinta.Styler()
            self.sep = sep

        def __str__(self):
            return self.s

        def __repr__(self):
            return self.__str__()

        def __eq__(self, other):
            if not isinstance(other, Tinta.Part):
                return False
            return self.s == other.s

        @property
        def has_formatting(self):
            return bool(self.styler.active_styles or self.styler.color_code)

        @property
        def color(self):
            return self.styler.color

        @property
        def color_code(self):
            return self.styler.color_code

        @property
        def style(self):
            return self.styler.active_styles

        @property
        def pln(self):
            return self.s

        def fmt(self, lp: "Optional[Tinta.Part]", np: "Optional[Tinta.Part]"):
            return stylize(self.s, lp, self, np)

        def esc(self, lp: "Optional[Tinta.Part]", np: "Optional[Tinta.Part]"):
            return esc(self.fmt(lp, np))

    @staticmethod
    def strip_ansi(s: str) -> str:
        """A utility method that strips ANSI escape codes from a string, converting a styled string into plaintext."""
        return re.sub(
            r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])", "", s, re.I | re.M | re.U
        )

    @classmethod
    def ljust(cls, s: str, width: int, fillchar: str = " ") -> str:
        """Returns a string left justified in a field of a specified width, accounting for ansi formatting."""
        chars_to_add = width - len(cls.strip_ansi(s))
        return f"{s}{str(fillchar or '') * chars_to_add}"

    @classmethod
    def rjust(cls, s: str, width: int, fillchar: str = " ") -> str:
        """Returns a string right justified in a field of a specified width, accounting for ansi formatting."""
        chars_to_add = width - len(cls.strip_ansi(s))
        return f"{str(fillchar or '') * chars_to_add}{s}"


def esc(string: str, replace: bool = False) -> str:
    """Returns the raw representation of a string. If replace is true,
    replace a double backslash with a single backslash."""
    r = repr(string)[1:-1]  # Strip the quotes from representation
    if replace:
        r = r.replace("\\\\", "\\")
    return r



from functools import reduce
from typing import cast, List, Sequence

from .multi_version_imports import TypeVar

T = TypeVar("T")


def flatmap(lst: Sequence[List[T]]) -> List[T]:
    """Flattens a list of lists into a single list. If the list is already flat,
    it is returned as is.
    """

    # if lst is not a nested list, return it as is
    if not any(isinstance(i, list) for i in lst):
        return cast(List[T], lst)

    return reduce(list.__add__, lst)


INDENT = 0


# def measure(func):

#     import tinta.constants as C

#     @wraps(func)
#     def timeit_wrapper(*args, **kwargs):
#         global INDENT
#         class_name = args[0].__class__.__name__ if args else ""
#         if class_name == "NoneType":
#             class_name = ""
#         n = f"{INDENT * ' '}{class_name} {func.__name__}"
#         print(n)
#         start_time = time.perf_counter()
#         INDENT += 2
#         result = func(*args, **kwargs)
#         end_time = time.perf_counter()
#         total_time = (end_time - start_time) * 1000
#         INDENT -= 2
#         # if total_time > 0.00001:
#         s = f"{n} - took {total_time:.2f} ms"
#         if total_time > 0.07:
#             col = "\x1b[38;5;196m"
#         elif total_time > 0.05:
#             col = "\x1b[38;5;208m"
#         elif total_time > 0.02:
#             col = "\x1b[38;5;220m"
#         else:
#             col = "\x1b[2;38;5;237m"
#         s = f"{col}{s}\x1b[0m" if col else s
#         print(s)
#         return result

#     return timeit_wrapper if C.PERF_MEASURE else func



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.


import os
import re
import time
import timeit
from contextlib import contextmanager
from typing import Any, Callable, Dict, List, Tuple

import pytest
from pytest import CaptureFixture

from tinta import Tinta

# pylint: disable=import-error
from tinta.stylize import _join, ansi_styles

Tinta.load_colors("examples/colors.ini")

O = "\x1b[0m"
GREEN = "\x1b[38;5;35m"
RED = "\x1b[38;5;1m"
BLUE = "\x1b[38;5;32m"
L_BLUE = "\x1b[38;5;37m"
YELLOW = "\x1b[38;5;214m"
AMBER = "\x1b[38;5;208m"
MINT = "\x1b[38;5;43m"
OLIVE = "\x1b[38;5;106m"
ORANGE = "\x1b[38;5;166m"
PURPLE = "\x1b[38;5;18m"
PINK = "\x1b[38;5;197m"
GRAY = "\x1b[38;5;243m"
D_GRAY = "\x1b[38;5;235m"
L_GRAY = "\x1b[38;5;248m"
WHITE = "\x1b[38;5;255m"
BOLD, BOLD_OFF = ansi_styles("bold")
UNDERLINE, UNDERLINE_OFF = ansi_styles("underline")
DIM, DIM_OFF = ansi_styles("dim")
STRIKE, STRIKE_OFF = ansi_styles("strikethrough")


def _stitch(*args):
    # get everything after \x1b[ for each arg
    codes = [re.search(r"\x1b\[(.*?)m", arg).group(1) for arg in args]
    return f"\x1b[{_join(*codes).strip(';')}m"


@contextmanager
def skip_on_github_actions():
    gh = bool(os.getenv("GITHUB_ACTIONS") == "true")
    if gh:
        pytest.mark.skipif(True, reason="Function skipped on GitHub Actions")
    yield gh


class TestInit:

    def test_init(self):
        assert len(Tinta("initialized").parts) == 1

    def test_accepts_string_on_init(self):
        assert Tinta("initialized").to_str() == "initialized"

    def test_load_colors(self, alt_colors_ini):
        # create a test colors.ini

        # load the colors
        Tinta.load_colors(alt_colors_ini)

        assert Tinta._colors.color_dict == {"sparkle": 195, "dragons_breath": 202}
        assert Tinta()._colors.color_list == ["sparkle", "dragons_breath"]

        # check the colors
        assert Tinta().sparkle("sparkle").to_str() == f"\x1b[38;5;195msparkle{O}"
        assert (
            Tinta().dragons_breath("dragon's breath").to_str()
            == f"\x1b[38;5;202mdragon's breath{O}"
        )


class TestBasicColorizing:

    basic_colors = [
        # fmt: off
        ("green", lambda: Tinta().green("green"), f"{GREEN}green{O}"),
        ("green", lambda: Tinta().tint(35, "green"), f"{GREEN}green{O}"),
        ("green", lambda: Tinta().tint("green", "green"), f"{GREEN}green{O}"),
        ("red", lambda: Tinta().red("red"), f"{RED}red{O}"),
        ("blue", lambda: Tinta().blue("blue"), f"{BLUE}blue{O}"),
        ("blue", lambda: Tinta().blue("green (but actually blue)"), f"{BLUE}green (but actually blue){O}"),
        ("light_blue", lambda: Tinta().light_blue("light_blue"), f"{L_BLUE}light_blue{O}"),
        ("yellow", lambda: Tinta().yellow("yellow"), f"{YELLOW}yellow{O}"),
        ("amber", lambda: Tinta().amber("amber"), f"{AMBER}amber{O}"),
        ("mint", lambda: Tinta().mint("mint"), f"{MINT}mint{O}"),
        ("olive", lambda: Tinta().olive("olive"), f"{OLIVE}olive{O}"),
        ("orange", lambda: Tinta().orange("orange"), f"{ORANGE}orange{O}"),
        ("purple", lambda: Tinta().purple("purple"), f"{PURPLE}purple{O}"),
        ("pink", lambda: Tinta().pink("pink"), f"{PINK}pink{O}"),
        ("gray", lambda: Tinta().gray("gray"), f"{GRAY}gray{O}"),
        ("dark_gray", lambda: Tinta().dark_gray("dark_gray"), f"{D_GRAY}dark_gray{O}"),
        ("light_gray", lambda: Tinta().light_gray("light_gray"), f"{L_GRAY}light_gray{O}"),
        ("white", lambda: Tinta().white("white"), f"{WHITE}white{O}"),
        # fmt: on
    ]

    @pytest.mark.parametrize(
        "color,Testa,expected",
        basic_colors,
    )
    def test_print_colors(self, color, Testa, expected, perf):
        Testa().print()

    @pytest.mark.parametrize(
        "color,Testa,expected",
        basic_colors,
    )
    def test_check_output(self, color, Testa, expected, perf):
        assert Testa().to_str() == expected

    def test_sandwich_colors(self):
        assert (
            Tinta("plain").green("green").default("plain").to_str()
            == f"plain {GREEN}green{O} plain"
        )

    def test_clear_resets_color(self):
        Tinta().green("green").clear("plain").print()
        assert (
            Tinta().green("green").clear("plain")("string").to_str()
            == f"{GREEN}green{O} plain string"
        )


class TestBasicStylizing:

    def test_bold(self):
        b = Tinta().bold("bold")
        assert b.to_str() == f"{BOLD}bold{O}"
        b.print()

        b = Tinta().b("bold")
        assert b.to_str() == f"{BOLD}bold{O}"
        b.print()

    def test_normal_resets_bold(self):
        b = Tinta().bold("bold").normal("normal")
        assert b.to_str() == f"{BOLD}bold{O} normal"
        b.print()

    def test_bold_resets_bold(self):
        b = Tinta("To").bold().red("boldly").bold().white("go")
        assert b.to_str() == f"To {_stitch(BOLD, RED)}boldly{O} {WHITE}go{O}"
        b.print()

    def test_bold_resets_only_bold(self):
        b = Tinta("To").bold().underline().red("boldly").bold().white("go")
        assert (
            b.to_str()
            == f"To {_stitch(BOLD, UNDERLINE, RED)}boldly{BOLD_OFF} {WHITE}go{O}"
        )
        b.print()

    def test_underline(self):
        u = Tinta().underline("underline")
        assert u.to_str() == f"{UNDERLINE}underline{O}"
        u.print()

        _ = Tinta()._("underscore")
        assert _.to_str() == f"{UNDERLINE}underscore{O}"
        _.print()

    def test_normal_resets_underline(self):
        u = Tinta().underline("underline").normal("normal")
        assert u.to_str() == f"{UNDERLINE}underline{O} normal"
        u.print()

    def test_underline_resets_underline(self):
        u = Tinta("To")._().amber("underlinely")._().purple("go")
        assert (
            u.to_str() == f"To {_stitch(UNDERLINE, AMBER)}underlinely{O} {PURPLE}go{O}"
        )
        u.print()

    def test_underline_resets_only_underline(self):
        u = Tinta("To").underline().bold().amber("underlinely").underline().olive("go")
        assert (
            u.to_str()
            == f"To {_stitch(BOLD, UNDERLINE, AMBER)}underlinely{UNDERLINE_OFF} {OLIVE}go{O}"
        )
        u.print()

    def test_dim(self):
        d = Tinta().dim("dim")
        assert d.to_str() == f"{DIM}dim{O}"
        d.print()

    def test_normal_resets_dim(self):
        t = Tinta().dim("dim").normal("normal")
        assert t.to_str() == f"{DIM}dim{O} normal"
        t.print()

    def test_dim_resets_dim(self):
        d = Tinta("To").dim().olive("dimly").dim().blue("go")
        assert d.to_str() == f"To {_stitch(DIM, OLIVE)}dimly{O} {BLUE}go{O}"
        d.print()

    def test_dim_resets_only_dim(self):
        d = Tinta("To").dim().underline().olive("dimly").dim().blue("go")
        assert (
            d.to_str()
            == f"To {_stitch(DIM, UNDERLINE, OLIVE)}dimly{DIM_OFF} {BLUE}go{O}"
        )
        d.print()

    def test_strikethrough(self):
        s = Tinta().strikethrough("strikethrough")
        assert s.to_str() == f"{STRIKE}strikethrough{O}"
        s.print()

    def test_normal_resets_strikethrough(self):
        s = Tinta().strikethrough("strikethrough").normal("normal")
        assert s.to_str() == f"{STRIKE}strikethrough{O} normal"
        s.print()

    def test_strikethrough_resets_strikethrough(self):
        s = (
            Tinta("To")
            .strikethrough()
            .olive("strikethroughly")
            .strikethrough()
            .blue("go")
        )
        assert (
            s.to_str() == f"To {_stitch(STRIKE, OLIVE)}strikethroughly{O} {BLUE}go{O}"
        )
        s.print()

    def test_strikethrough_resets_only_strikethrough(self):
        s = (
            Tinta("To")
            .strikethrough()
            .underline()
            .olive("strikethroughly")
            .strikethrough()
            .blue("go")
        )
        assert (
            s.to_str()
            == f"To {_stitch(UNDERLINE, STRIKE, OLIVE)}strikethroughly{STRIKE_OFF} {BLUE}go{O}"
        )
        s.print()

    def test_bold_underline(self):
        bu = Tinta().bold("bold").underline("underline")
        assert bu.to_str() == f"{BOLD}bold {UNDERLINE}underline{O}"
        bu.print()

    def test_bold_dim(self):
        bd = Tinta().bold("bold").dim("dim")
        assert bd.to_str() == f"{BOLD}bold {DIM}dim{O}"
        bd.print()

    def test_underline_dim(self):
        ud = Tinta().underline("underline").dim("dim")
        assert ud.to_str() == f"{UNDERLINE}underline {DIM}dim{O}"
        ud.print()

    def test_bold_underline_dim(self):
        bud = Tinta().bold("bold").underline("underline").dim("dim")
        assert bud.to_str() == f"{BOLD}bold {UNDERLINE}underline {DIM}dim{O}"
        bud.print()

    def test_underline_bold_dim(self):
        ubd = Tinta().underline("underline").bold("bold").dim("dim")
        assert ubd.to_str() == f"{UNDERLINE}underline {BOLD}bold {DIM}dim{O}"
        ubd.print()

    def test_bold_strikethrough(self):
        bs = Tinta().bold("bold").strikethrough("strikethrough")
        assert bs.to_str() == f"{BOLD}bold {STRIKE}strikethrough{O}"
        bs.print()

    def test_underline_strikethrough(self):
        us = Tinta().underline("underline").strikethrough("strikethrough")
        assert us.to_str() == f"{UNDERLINE}underline {STRIKE}strikethrough{O}"
        us.print()

    def test_bold_underline_strikethrough(self):
        bus = Tinta().bold("bold").underline("underline").strikethrough("strikethrough")
        assert (
            bus.to_str() == f"{BOLD}bold {UNDERLINE}underline {STRIKE}strikethrough{O}"
        )
        bus.print()

    def test_styles_remain_until_normal(self):
        assert (
            Tinta()
            .bold("bold")
            .underline("underline")
            .dim("dim")
            .pink("pink")
            .normal()
            .pink("pink")
            .to_str()
            == f"{BOLD}bold {UNDERLINE}underline {DIM}dim {PINK}pink{_stitch(BOLD_OFF, UNDERLINE_OFF)} pink{O}"
        )


class TestChaining:

    @pytest.mark.parametrize(
        "Testa, kwargs, expected",
        [
            # fmt: off
            (lambda: Tinta().green("green").red("red").blue("blue"), {"sep":"", "plaintext":True}, "greenredblue"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {"plaintext": True}, "green red blue"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {"sep": ""}, f"{GREEN}green{RED}red{BLUE}blue{O}"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {}, f"{GREEN}green {RED}red {BLUE}blue{O}"),
            (lambda: Tinta().green("green").red("red").blue("blue "), {}, f"{GREEN}green {RED}red {BLUE}blue{O} "),
            (lambda: Tinta().green("green").red("red").blue("blue \n"), {}, f"{GREEN}green {RED}red {BLUE}blue{O} \n"),
            (lambda: Tinta().green("green").red("red").blue("blue").clear(), {}, f"{GREEN}green {RED}red {BLUE}blue{O}"),
            # fmt: on
        ],
    )
    def test_chaining_resets_correctly(
        self,
        Testa: Callable,
        kwargs: Dict[str, Any],
        expected: str,
        capfd: CaptureFixture[str],
    ):
        s = Testa().to_str(**kwargs)
        Testa().print(**kwargs)
        assert s == expected
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            (
                lambda: Tinta().mint("Mint\nice cream"),
                f"{MINT}Mint\nice cream{O}",
            ),
            (
                lambda: Tinta().mint("Mint").white("\nice cream\nis the")._("\nbest"),
                f"{MINT}Mint{WHITE}\nice cream\nis the{UNDERLINE}\nbest{O}",
            ),
            (
                lambda: Tinta("\n").mint("Mint\n").white("ice cream is the")._("best"),
                f"\n{MINT}Mint\n{WHITE}ice cream is the {UNDERLINE}best{O}",
            ),
        ],
    )
    def test_newlines_in_chaining(self, Testa: Callable[[], Tinta], expected: str):
        s = Testa().to_str()
        Testa().print()
        assert s == expected

    def test_chaining_same_styles_doesnt_dupe(self):
        t = Tinta().green("green").green("green").green("green")
        assert t.to_str() == f"{GREEN}green green green{O}"
        t.print()


class TestUnicode:

    def test_unicode(self):
        Tinta().pink("I ♡ Unicorns").print()
        assert Tinta().pink("I ♡ Unicorns").to_str() == f"{PINK}I ♡ Unicorns{O}"

    def test_box_drawing(self):
        s = f"┌{'─'*14}┐\n│ I ♡ Unicorns │\n└{'─'*14}┘"
        Tinta("\n").mint(s).print()
        assert (
            Tinta().mint(s).to_str()
            == f"{MINT}┌──────────────┐\n│ I ♡ Unicorns │\n└──────────────┘{O}"
        )

    def test_emoji(self):
        Tinta("🦄").print()
        assert Tinta("🦄").to_str() == "🦄"

    def test_emoji_with_color(self):
        t = Tinta().purple("🦄")
        assert Tinta().purple("🦄").to_str() == f"{PURPLE}🦄{O}"
        t.print()

    def test_emoji_with_color_and_clear(self):
        t = Tinta().pink("🦄").clear("🦄")
        assert t.to_str(sep="") == f"{PINK}🦄{O}🦄"
        t.print(sep="")

    def test_emoji_with_color_style_and_normal(self):
        t = Tinta().purple()._("🦄").normal("🦄").pink("🦄")
        assert (
            t.to_str()
            == f"{_stitch(UNDERLINE, PURPLE)}🦄{UNDERLINE_OFF} 🦄 {PINK}🦄{O}"
        )
        t.print()

    def test_emoji_with_color_style_and_clear(self):
        t = Tinta().purple()._("🦄").clear("🦄").pink("🦄")
        assert t.to_str() == f"{_stitch(UNDERLINE, PURPLE)}🦄{O} 🦄 {PINK}🦄{O}"
        t.print()


class TestLowerLevel:

    ITS_NOT_EASY = "it's not easy being green"

    def assert_its_not_easy_being_green(self, out: str):
        assert self.ITS_NOT_EASY in out
        assert f"{GREEN}" in out

    def test_tint_takes_int_arg0_as_color(self):
        out = Tinta().tint(35, self.ITS_NOT_EASY).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_str_arg0_as_color(self):
        out = Tinta().tint("green", self.ITS_NOT_EASY).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_int_color_kwarg(self):
        out = Tinta().tint(self.ITS_NOT_EASY, color=35).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_str_color_kwarg(self):
        out = Tinta().tint(self.ITS_NOT_EASY, color="green").to_str()
        self.assert_its_not_easy_being_green(out)


class TestEdgeCases:

    def test_missing_color(self):
        with pytest.raises(Exception):
            Tinta().sparkle().print()

    def test_color_same_as_builtin_method(self):
        with pytest.raises(Exception):
            Tinta().load_colors("tests/test_colors_invalid.ini")

    def test_print_empty(self):
        Tinta().print()

    def test_print_empty_str(self):
        Tinta("").print()

    def test_print_none(self):
        Tinta(None).print()

    def test_empty_color_call(self):
        t = Tinta("Plain").green()
        t.push("Green")
        t.print()

    def test_tint_color_0(self):
        Tinta().tint(0, "Zero").print()


class TestExamples:

    def test_basic_example(self):
        from examples.basic_example import basic

        Tinta._initialized = False
        Tinta._colors._initialized = False

        basic()

    def test_complete_example(self):
        from examples.complete_example import complete

        Tinta._initialized = False
        Tinta._colors._initialized = False

        complete()


class TestWhiteSpace:

    def test_print_whitespace(self):
        Tinta(" ").print()

    whitespace_cases = [
        (lambda: Tinta("Hello\nWorld"), "Hello\nWorld"),
        (lambda: Tinta("\nHello\nWorld"), "\nHello\nWorld"),
        (lambda: Tinta("\nHello\nWorld\n"), "\nHello\nWorld\n"),
        (lambda: Tinta("\nHello\nWorld\n\n"), "\nHello\nWorld\n\n"),
        (lambda: Tinta("Hello\nWorld\n"), "Hello\nWorld\n"),
        (lambda: Tinta("Hello World\n"), "Hello World\n"),
        (lambda: Tinta("Hello World"), "Hello World"),
        (lambda: Tinta("Hello World "), "Hello World "),
        (lambda: Tinta(" Hello World "), " Hello World "),
        (lambda: Tinta(" Hello World"), " Hello World"),
        (lambda: Tinta("  Hello World"), "  Hello World"),
        (lambda: Tinta("  Hello World "), "  Hello World "),
        (lambda: Tinta("  Hello World  "), "  Hello World  "),
        (lambda: Tinta(" Hello World\n"), " Hello World\n"),
        (lambda: Tinta(" Hello World\n "), " Hello World\n "),
        (lambda: Tinta(" Hello World\n  "), " Hello World\n  "),
        (lambda: Tinta(" Hello World\n\n"), " Hello World\n\n"),
        (lambda: Tinta(" Hello World\n\n "), " Hello World\n\n "),
        (lambda: Tinta("\nHello World \n"), "\nHello World \n"),
        (lambda: Tinta("\nHello World \n "), "\nHello World \n "),
        (lambda: Tinta("\n Hello World \n\n"), "\n Hello World \n\n"),
        (lambda: Tinta("\n Hello\n World \n"), "\n Hello\n World \n"),
    ]

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_uncolored_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        assert s == expected
        Testa().print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_uncolored_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        assert p == expected
        Testa().print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_to_str_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        found = re.search(r"\s+$", expected)
        trail_ws = found.group() if found else ""
        _expected = f"{PURPLE}{expected.rstrip()}{O}{trail_ws}"
        #                                         ^^^^^^^^^^^^^
        # Reset ANSI char is always placed before trailing whitespace
        assert Tinta().purple(s).to_str() == _expected
        Tinta().purple(s).print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{_expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_to_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        assert Tinta().purple(p).to_str(plaintext=True) == expected
        Tinta().purple(p).print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_chain_to_str_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        found = re.search(r"\s+$", expected)
        trail_ws = found.group() if found else ""
        expected_purple = f"{PURPLE}{expected}"
        expected_pink = f"{PINK}{expected.rstrip()}{O}{trail_ws}"
        l_or_r_newline = expected.startswith("\n") or expected.endswith("\n")
        sep = "" if l_or_r_newline else " "
        # ^ If there is a newline at the end of the current or start of the next segment,
        #   Tinta will ignore the separator
        assert (
            Tinta().purple(s).pink(s).to_str()
            == f"{expected_purple}{sep}{expected_pink}"
        )
        Tinta().purple(s).pink(s).print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected_purple}{sep}{expected_pink}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_chain_to_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        l_or_r_newline = expected.startswith("\n") or expected.endswith("\n")
        sep = "" if l_or_r_newline else " "
        # ^ If there is a newline at the end of the current or start of the next segment,
        #   Tinta will ignore the separator
        assert (
            Tinta().purple(p).pink(p).to_str(plaintext=True)
            == f"{expected}{sep}{expected}"
        )
        Tinta().purple(p).pink(p).print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}{sep}{expected}\n"

    def test_sep_inherits_prev_part_color(self):
        assert (
            Tinta().red("Red").green("Green").blue("Blue").to_str(sep=";")
            == f"{RED}Red;{GREEN}Green;{BLUE}Blue{O}"
        )


class TestMultipleInstances:

    def test_same_instance_chains_to_str(self):
        t = Tinta()
        red = f"{RED}Red"
        green = f"{GREEN}Green"
        blue = f"{BLUE}Blue"
        assert t.red("Red").to_str() == f"{red}{O}"
        assert t.green("Green").to_str() == f"{red} {green}{O}"
        assert t.blue("Blue").to_str() == f"{red} {green} {blue}{O}"

    def test_same_instance_resets_after_print(self):
        t = Tinta()
        assert t.red("Red").to_str() == f"{RED}Red{O}"
        t.print()
        assert t.green("Green").to_str() == f"{GREEN}Green{O}"
        t.print()
        assert t.blue("Blue").to_str() == f"{BLUE}Blue{O}"

    def test_multiple_instances_dont_interfere(self):
        t1 = Tinta()
        t2 = Tinta()
        t3 = Tinta()
        assert t1.red("Red").to_str() == f"{RED}Red{O}"
        assert t2.green("Green").to_str() == f"{GREEN}Green{O}"
        assert t3.blue("Blue").to_str() == f"{BLUE}Blue{O}"


# @pytest.mark.skip
class TestFeatures:

    def test_inspect(self):
        assert Tinta().inspect(code=0) == "default"
        assert Tinta().inspect(name="default") == 0

        assert Tinta().inspect(code=35) == "green"
        assert Tinta().inspect(name="green") == 35

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            (lambda: Tinta("Hello")("world"), "Hello world"),
            (lambda: Tinta("Hello ")("world"), "Hello  world"),
            (lambda: Tinta("Hello")(", world"), "Hello, world"),
            (lambda: Tinta("Hello,")("world"), "Hello, world"),
            (lambda: Tinta("Hello")(",")("world"), "Hello, world"),
            (lambda: Tinta("Hello world")("!"), "Hello world!"),
            (lambda: Tinta("Nice")("."), "Nice."),
            (lambda: Tinta("Nice")(".gif file"), "Nice .gif file"),
            (lambda: Tinta(" *** Important"), " *** Important"),
            (lambda: Tinta(", world"), ", world"),
            (lambda: Tinta(" , world"), " , world"),
            (lambda: Tinta(",")("world"), ", world"),
            (lambda: Tinta(" ,")("world"), " , world"),
            (lambda: Tinta(" ,")(" world"), " ,  world"),
            (
                lambda: Tinta()
                .red(" *** Important : Smart fix punctuation should work with")
                .green("ANSI")
                .red(", as well as plaintext."),
                f"{RED} *** Important : Smart fix punctuation should work with {GREEN}ANSI{RED}, as well as plaintext.{O}",
            ),
        ],
    )
    def test_smart_fix_punctuation(self, Testa: Callable[[], Tinta], expected: str):
        assert Testa().to_str() == expected


class TestComplexStructure:

    @pytest.mark.parametrize(
        "inputs,expected",
        [
            (
                (
                    [
                        ("How long", None),
                        ("can two people talk about nothing?", None),
                        ("Hmm?", None),
                    ],
                    "How long can two people talk about nothing? Hmm?",
                )
            ),
            (
                (
                    [
                        ("How long", ""),
                        ("can two people talk about nothing?", ""),
                        ("Hmm?", ""),
                    ],
                    "How longcan two people talk about nothing?Hmm?",
                )
            ),
            (
                (
                    [
                        ("How long", " "),
                        ("can two people", " "),
                        ("talk about nothing?", " "),
                    ],
                    "How long can two people talk about nothing?",
                )
            ),
            (
                (
                    [("A section", " "), ("of text", ""), (", separated.", " ")],
                    "A section of text, separated.",
                )
            ),
        ],
    )
    def test_sep(self, inputs: List[Tuple[str, str]], expected: str) -> None:
        t = Tinta()
        for s, sep in inputs:
            t.push(s, sep=sep)
        assert t.to_str(plaintext=True) == expected

    def testparts(self):
        assert len(Tinta().green("green").red("red").blue("blue").parts) == 3
        assert (
            len(
                Tinta()
                .green("green")
                .underline()
                .red("red")
                .blue("blue")
                .yellow("yellow")
                .parts_fmt
            )
            == 4
        )
        assert (
            len(Tinta().green("green").underline().red("red").blue("blue").parts_pln)
            == 3
        )

    def test_f_strings(self):
        dog = "cat"
        assert (
            Tinta(f"A {dog} is a human's best friend").to_str(plaintext=True)
            == "A cat is a human's best friend"
        )
        assert Tinta(f"A {Tinta().red('hologram').to_str()} is a human's best friend")

    def test_push(self):
        t = Tinta().push("How long").push("can two people talk about nothing?")
        assert t.to_str(plaintext=True) == "How long can two people talk about nothing?"

        assert len(t.parts) == 2
        assert len(t.parts_fmt) == 2
        assert len(t.parts_pln) == 2

        t = Tinta().push("How long", "can two people talk about nothing?")
        assert t.to_str(plaintext=True) == "How long can two people talk about nothing?"

        assert len(t.parts) == 1
        assert len(t.parts_fmt) == 1
        assert len(t.parts_pln) == 1

    @pytest.mark.parametrize(
        "inputs,expected",
        [
            (
                ("How long", "can two people talk about nothing?"),
                "How long can two people talk about nothing?",
            ),
            (
                (["How long", "can two people talk about nothing?"]),
                "How long can two people talk about nothing?",
            ),
        ],
    )
    def test_to_str(self, inputs: Tuple[str], expected):
        t = Tinta().push(*inputs)
        assert t.to_str(plaintext=True) == expected
        assert t.to_str(plaintext=True) == expected

    def test_pop(self):
        t = Tinta().push("How long").push("can two people talk about nothing?")
        t.pop()
        assert t.to_str(plaintext=True) == "How long"
        assert len(t.parts) == 1
        assert len(t.parts_fmt) == 1
        assert len(t.parts_pln) == 1

        t = Tinta()
        for p in range(10):
            t.push(str(p))

        assert len(t.parts) == 10
        t.pop(10)

        assert len(t.parts) == 0

        t.pop(2)  # Shouldn't error if we remove more than we have

        assert len(t.parts) == 0

    def test_pop_updates_current_style(self):
        t = Tinta().green("green").red("red").blue("blue")
        t.pop()
        assert t.to_str(plaintext=True) == "green red"
        t.pop()
        assert t.to_str(plaintext=True) == "green"
        t.pop()
        assert t.to_str(plaintext=True) == ""

        t = Tinta().green("green").red("red").blue("blue").bold("bold n' blue")
        t.pop(2)
        assert t._styler == Tinta.Styler(color=RED)

    def test_tint_handles_zero(self):
        s = (
            Tinta("white")
            .red("red")
            .green("green")
            .blue("blue")
            .tint(0, "default")
            .to_str(sep=" ")
        )
        print(s)
        assert s == f"white {RED}red {GREEN}green {BLUE}blue{O} default"


class TestUtils:

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            # fmt: off
            (lambda: Tinta().blue("blue"), "blue"),
            (lambda: Tinta().red("red"), "red"),
            (lambda: Tinta().green("green"), "green"),
            (lambda: Tinta().yellow("yellow"), "yellow"),
            (lambda: Tinta().pink("🦄"), "🦄"),
            # fmt: on
            (
                lambda: Tinta()
                .pink("Pink unicorns 🦄")
                .purple("are as good as purple 🦄 ones"),
                "Pink unicorns 🦄 are as good as purple 🦄 ones",
            ),
            (
                lambda: Tinta()
                .purple("Hello purple world!")
                .clear("I've got a song to sing about")
                ._("unicorns 🦄"),
                "Hello purple world! I've got a song to sing about unicorns 🦄",
            ),
        ],
    )
    def test_strip_ansi(self, Testa: Callable, expected: str):
        assert Tinta.strip_ansi(Testa().to_str()) == expected

    @pytest.mark.parametrize(
        "Testa, ljust, fillchar, expected",
        [
            # fmt: off
            (lambda: Tinta().mint("🦄"), 10, " ", f"{MINT}🦄{O}         "),
            (lambda: Tinta().mint("I love unicorns and magical fruit"), 10, " ", f"{MINT}I love unicorns and magical fruit{O}"),
            (lambda: Tinta("ljust"), 13, "-", f"ljust{'-'*8}"),
            # fmt: on
            (
                lambda: Tinta("Wild")
                .purple("unicorns")
                .pink("like their")
                .mint("space"),
                20,
                " ",
                f"Wild {PURPLE}unicorns {PINK}like their {MINT}space{O}",
            ),
            (
                lambda: Tinta("a").purple("b").pink("c d").mint("e "),
                80,
                " ",
                f"a {PURPLE}b {PINK}c d {MINT}e{O} {' '*70}",
            ),
        ],
    )
    def test_ljust(self, Testa: Callable, ljust: int, fillchar: str, expected: str):
        assert Tinta.ljust(Testa().to_str(), ljust, fillchar) == expected

    @pytest.mark.parametrize(
        "Testa, rjust, fillchar, expected",
        [
            # fmt: off
            (lambda: Tinta().mint("🦄"), 10, " ", f"         {MINT}🦄{O}"),
            (lambda: Tinta().mint("I love unicorns and magical fruit"), 10, " ", f"{MINT}I love unicorns and magical fruit{O}"),
            (lambda: Tinta("rjust"), 13, "-", f"{'-'*8}rjust"),
            # fmt: on
            (
                lambda: Tinta("Wild")
                .purple("unicorns")
                .pink("like their")
                .mint("space"),
                20,
                " ",
                f"Wild {PURPLE}unicorns {PINK}like their {MINT}space{O}",
            ),
            (
                lambda: Tinta(" a").purple("b").pink("c d").mint("e"),
                80,
                " ",
                f"{' '*70} a {PURPLE}b {PINK}c d {MINT}e{O}",
            ),
        ],
    )
    def test_rjust(self, Testa: Callable, rjust: int, fillchar: str, expected: str):
        assert Tinta.rjust(Testa().to_str(), rjust, fillchar) == expected


class TestTiming:

    def test_basic_time_against_raw(self):
        base_start = time.perf_counter()
        print(
            f"{BLUE}blue {GREEN}green {L_BLUE}light blue {YELLOW}yellow {AMBER}amber {MINT}mint {OLIVE}olive {ORANGE}orange {PURPLE}purple {PINK}pink {GRAY}gray {D_GRAY}dark gray {L_GRAY}light gray {WHITE}white{O}"
        )
        base_end = time.perf_counter()

        tinta_start = time.perf_counter()

        (
            Tinta()
            .blue("blue")
            .green("green")
            .light_blue("light blue")
            .yellow("yellow")
            .amber("amber")
            .mint("mint")
            .olive("olive")
            .orange("orange")
            .purple("purple")
            .pink("pink")
            .gray("gray")
            .dark_gray("dark gray")
            .light_gray("light gray")
            .white("white")
            .print()
        )
        tinta_end = time.perf_counter()

        raw_diff = base_end - base_start
        tinta_diff = tinta_end - tinta_start

        print(f"Raw: {raw_diff:.4f} seconds")
        print(f"Tinta: {tinta_diff:.4f} seconds")

        assert tinta_diff - raw_diff < 0.001

    def test_with_timeit(self):

        count = 500

        tinta_time = timeit.timeit(
            lambda: Tinta()
            .blue("blue")
            .green("green")
            .light_blue("light blue")
            .yellow("yellow")
            .amber("amber")
            .mint("mint")
            .olive("olive")
            .orange("orange")
            .purple("purple")
            .pink("pink")
            .gray("gray")
            .dark_gray("dark gray")
            .light_gray("light gray")
            .white("white")
            .print(),
            number=count,
        )

        class FakeClass:
            def __init__(self):
                self.blue = "blue"
                self.green = "green"
                self.light_blue = "light blue"
                self.yellow = "yellow"
                self.amber = "amber"
                self.mint = "mint"
                self.olive = "olive"
                self.orange = "orange"
                self.purple = "purple"
                self.pink = "pink"
                self.gray = "gray"
                self.dark_gray = "dark gray"
                self.light_gray = "light gray"
                self.white = "white"

            def print(self):
                print(
                    f"{BLUE}{self.blue} {GREEN}{self.green} {L_BLUE}{self.light_blue} {YELLOW}{self.yellow} {AMBER}{self.amber} {MINT}{self.mint} {OLIVE}{self.olive} {ORANGE}{self.orange} {PURPLE}{self.purple} {PINK}{self.pink} {GRAY}{self.gray} {D_GRAY}{self.dark_gray} {L_GRAY}{self.light_gray} {WHITE}{self.white}{O}"
                )

        class_time = timeit.timeit(
            lambda: FakeClass().print(),
            number=count,
        )

        raw_time = timeit.timeit(
            lambda: print(
                f"{BLUE}blue {GREEN}green {L_BLUE}light blue {YELLOW}yellow {AMBER}amber {MINT}mint {OLIVE}olive {ORANGE}orange {PURPLE}purple {PINK}pink {GRAY}gray {D_GRAY}dark gray {L_GRAY}light gray {WHITE}white{O}"
            ),
            number=count,
        )

        print(f"Raw: {raw_time:.4f} seconds")
        print(f"Class: {class_time:.4f} seconds")
        print(f"Tinta: {tinta_time:.4f} seconds")

        assert tinta_time - raw_time < 0.5
